"""A detectron2 hook system to make validation during the training process."""

from detectron2.engine.hooks import HookBase
from detectron2.utils.logger import log_every_n_seconds
import detectron2.utils.comm as comm
import torch
import time
from datetime import datetime
import numpy as np
import torch
import time
import datetime
import logging


class LossEvalHook(HookBase):
    def __init__(self, eval_period, model, data_loader):
        """_summary_

        Args:
            eval_period (int): the number of iteration needed to perform a validation step.
            model: a neural network
            data_loader (:class:`torch.utils.data.DataLoader`): a torch dataloader
        """
        self._model = model
        self._period = eval_period
        self._data_loader = data_loader

    def _do_loss_eval(self):
        # Copying inference_on_dataset from evaluator.py
        total = len(self._data_loader)
        num_warmup = min(5, total - 1)

        start_time = time.perf_counter()
        total_compute_time = 0
        losses = []
        for idx, inputs in enumerate(self._data_loader):
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_compute_time = 0
            start_compute_time = time.perf_counter()
            if torch.cuda.is_available():
                torch.cuda.synchronize()
            total_compute_time += time.perf_counter() - start_compute_time
            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            seconds_per_img = total_compute_time / iters_after_start
            if idx >= num_warmup * 2 or seconds_per_img > 5:
                total_seconds_per_img = (time.perf_counter() - start_time) / iters_after_start
                eta = datetime.timedelta(seconds=int(total_seconds_per_img * (total - idx - 1)))
                log_every_n_seconds(
                    logging.INFO,
                    "Loss on Validation  done {}/{}. {:.4f} s / img. ETA={}".format(
                        idx + 1, total, seconds_per_img, str(eta)
                    ),
                    n=5,
                )
            loss_batch = self._get_loss(inputs)
            losses.append(loss_batch)
        mean_loss = np.mean(losses)
        # self.trainer.storage.put_scalar('validation_loss', mean_loss)
        comm.synchronize()

        # return losses
        return mean_loss

    def _get_loss(self, data):
        # How loss is calculated on train_loop
        metrics_dict = self._model(data)
        metrics_dict = {
            k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else float(v)
            for k, v in metrics_dict.items()
        }
        total_losses_reduced = sum(loss for loss in metrics_dict.values())
        return total_losses_reduced
        
        
    def after_step(self):
        next_iter = int(self.trainer.iter) + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            mean_loss = self._do_loss_eval()
            self.trainer.storage.put_scalars(validation_loss=mean_loss)
            print("validation do loss eval", mean_loss)
        else:
            pass
            # self.trainer.storage.put_scalars(timetest=11)


"""A module to convert a :class:`fiftyone.core.dataset` into a geojson file.

Annotations in the fiftyone dataset are masks or bounding box, mapped as pixel values 
but if they match with a georeferenced raster file (a tif file for example), 
they are spatialized with an affine transformation.

Typical usage examples:

    dataset = fo.load_dataset('dataset_name')

    geojson_exporter = GeoFiftyoneExporter(
    export_dir = '/path/to/directory/', 
    epsg_code = 'EPSG:4326', 
    label_type = 'detections',
    dest_name = 'spatial_predictions.geojson'
    )

    dataset.export(
    dataset_exporter=geojson_exporter,
    label_field='predictions',
    export_media=False
    )

"""

import pandas as pd
import fiftyone as fo
import fiftyone.utils.data as foud
from shapely import Point
import geopandas
import os
from artus.spatialize.GeoCOCOExporter import GeoCOCOExporter

class GeoFiftyoneExporter(foud.LabeledImageDatasetExporter, GeoCOCOExporter): 
    """Export a fiftyone dataset to a geospatial format (geojson).

     This is only possible if samples are raster format (i.e. tif).

     Datasets of this type are exported in the following format:

         <export_dir>/
             dest_name.geojson

     where ``dest_name.geojson`` is a GeoJson file containing labels.
    
    Attributes:
         export_dir: the directory to write the export
         label_type : the label_type of the concerned fiftyone field ('polylines' 
            for segmentation annotations or 'detections' for bbox annotations)
         epsg_code : the epsg code (for example : '4326' for world coordinates) in which the results will be exported
         dest_name : the file name of the geojson file with the extension (for example : 'spatial_predictions.geojson')
    """

    def __init__(self, export_dir, label_type, epsg_code, dest_name):
        super().__init__(export_dir=export_dir)
        self._data_dir = None
        self._labels_path = None
        self._labels = None
        self._image_exporter = None
        self.label_type = label_type
        self.epsg_code = epsg_code
        self.dest_name = dest_name
        self.sample_dir = None

    @property
    def requires_image_metadata(self):
        """Whether this exporter requires
         :class:`fiftyone.core.metadata.ImageMetadata` instances for each sample
         being exported.
        """
        return True

    @property
    def label_cls(self):
        """The :class:`fiftyone.core.labels.Label` class(es) exported by this
         exporter.

         This can be any of the following:

         -   a :class:`fiftyone.core.labels.Label` class. In this case, the
             exporter directly exports labels of this type
         -   a list or tuple of :class:`fiftyone.core.labels.Label` classes. In
             this case, the exporter can export a single label field of any of
             these types
         -   a dict mapping keys to :class:`fiftyone.core.labels.Label` classes.
             In this case, the exporter can handle label dictionaries with
             value-types specified by this dictionary. Not all keys need be
             present in the exported label dicts
         -   ``None``. In this case, the exporter makes no guarantees about the
             labels that it can export
        """
        return [fo.Detections, fo.Polylines]

    def setup(self):
        """Performs any necessary setup before exporting the first sample in
         the dataset.

         This method is called when the exporter's context manager interface is
         entered, :func:`DatasetExporter.__enter__`.
        """
        self._labels_path = os.path.join(self.export_dir, self.dest_name)
        self._labels = []

        self.columns_names = ['img_filename', 'label' ,'confidence', 'geometry']
        
        self._image_exporter = foud.ImageExporter(
            False, export_path=self._data_dir, default_ext=".jpg",
        )
        self._image_exporter.setup()

    def log_collection(self, sample_collection):
        self.sample_collection = sample_collection

    def shapely_polygons(self, label, metadata):
        """Transform a :class:`fiftyone.core.label.Label` into 
        a shapely polygon. 

        Args:
            label (:class:`fiftyone.core.label.Label`): a detection or polylines field.
            metadata (:class:`fiftyone.core.metadata.ImageMetadata`): metadata of the sample

        Returns:
            :class:`shapely.Polygon`: a shapely polygon that is a square if label was 
            bouding boxes or a polygon if label was a polyline. Shapely polygons are still pixel values.
        """

        imw, imh = (metadata.width, metadata.height)
        if self.label_type=='detections':
            polygon = label.to_shapely(frame_size=(imw,imh))
        elif self.label_type=='polylines':
            polygon = label.to_shapely(frame_size=(imw,imh))
        return polygon

    def export_sample(self, image_or_path, label, metadata=None):
        """Exports the given sample to the dataset.
            
        If images in the dataset are tif then it will convert use the world coordinates for the masks or bbox...
        Otherwise, if images are simply georeferenced it will assigned every label on the image to a single GPS point
        GPS point must have been added to the dataset using :func:`artus.spatialize.LocationImporter.import_csv_locations`

        Args:
            image_or_path: an image or the path to the image on disk
            label: an instance of :meth:`label_cls`, or a dictionary mapping
                 field names to :class:`fiftyone.core.labels.Label` instances,
                 or ``None`` if the sample is unlabeled
            metadata (None): a :class:`fiftyone.core.metadata.ImageMetadata`
                 instance for the sample. Only required when
                 :meth:`requires_image_metadata` is ``True``
        """

        sample = self.sample_collection[image_or_path]
        metadata = sample.metadata

        # Get field values
        for n_label in getattr(label, self.label_type):

            img_filename = sample.filepath
            label = n_label.label
            confidence = n_label.confidence
            
            if metadata.mime_type == "image/tiff":
                geometry = self.shapely_polygons(n_label, metadata) 
                   
            elif metadata.mime_type == "image/tiff" and self.sample_collection.has_field('location'):
                geometry = Point(sample.location.point)

            gdf_row = (img_filename , label , confidence, geometry)

            self._labels.append(gdf_row)
        
        
    def close(self, *args):
        """Performs any necessary actions after the last sample has been
        exported.
        This method is called when the exporter's context manager interface is
        exited, :func:`DatasetExporter.__exit__`. Polygons are converted to a world
        coordinates with an affine transformation.

        Args:
            *args: the arguments to :func:`DatasetExporter.__exit__`
        """
        
        # Ensure the base output directory exists
        if not os.path.exists(self.export_dir):
            os.makedirs(self.export_dir)
            
        #convert _labels into gdf
        df = pd.DataFrame(data=self._labels, columns=self.columns_names)
        gdf = geopandas.GeoDataFrame(df, geometry='geometry')

        #convert pixel-values coordinates into geospatial coordinates
        if gdf['geometry'].geom_type.all() != 'Point':
            gdf['transform'] =  [self.get_transform(sample) for sample in gdf['img_filename']]
            gdf = self.affine_transform(gdf)        

        gdf = gdf.filter(items=['img_filename', 'label', 'confidence', 'geometry'])
        gdf.to_file(os.path.join(self.export_dir, self.dest_name), driver='GeoJSON')


"""A module to import sample location from a CSV.

This module is dedicated to user that only have images georeferenced by a GPS point. It
will add the matching GPS point to every sample in the fiftyone dataset."""

import fiftyone as fo
import pandas as pd
import tqdm
import os

def import_csv_locations(location_path, fiftyone_dataset):
    """A function to import GPS locations from a CSV file and add it to  the matching sample in a fiftyone dataset.

    Args:
        location_path (str): a path to csv containing 3 columns called 'filename', 'latitude' and 'longitude'
        fiftyone_dataset (:class:`fiftyone.core.dataset`): a fiftyone dataset containing images

    Returns:
        :class:`fiftyone.core.dataset`: the same dataset with a geolocation point added to the samples listed in the csv file.
    """
    locations_df = pd.read_csv(location_path)

    for sample in tqdm(fiftyone_dataset):
        image_location = locations_df[locations_df['filename']==os.path.basename(sample['filepath'])]
        sample['location'] = fo.GeoLocation(point=[image_location['latitude'].iloc[0], image_location['longitude'].iloc[0]])
        sample.save()

    fiftyone_dataset.save()
    return fiftyone_dataset




"""Module to use a model checkpoint and predict annotations on unlabeled images.

Annotations produced by the model (called predictions) are segmentation masks for 
instance segmentation tasks or bounding boxes for object detection tasks.
"""

from detectron2.engine import DefaultPredictor
from detectron2.structures import Boxes, Instances
from torchvision.transforms import functional as func
import numpy as np
import cv2
import fiftyone as fo
import fiftyone.utils.labels as foul
from fiftyone import ViewField as F
import os
import cv2
from torchvision.ops import nms
import fiftyone as fo
from artus.inference.config import add_config
from joblib import Parallel, delayed
import multiprocessing
os.environ["OPENCV_IO_MAX_IMAGE_PIXELS"]=pow(2,50).__str__()

def build_predictor(config_path, device):
    """ Load a model in indeference mode.
    
    Args:
        config_path (str): the path to a config file in yaml format
        device (str): 'cpu' or 'cuda'. Results of `"cuda" if torch.cuda.is_available() else "cpu"`
    
    Returns 
        predictor : the model loaded to predict unlabeled images
    """
    cfg = add_config(config_path, device)
    predictor = DefaultPredictor(cfg)
    return predictor

def predict(predictor, filepath):
    """Predict masks or bounding boxes on an image.

    Args:
        predictor : a detectron2 predictor returned by `build_predictor()`
        filepath (str): the absolute or relative path to the image
    
    Returns:
        an ordered dict with the label predicted and the coordinates of the bbox or masks
    """
    image = cv2.imread(filepath)
    outputs = predictor(image)
    return outputs

def crop_mask(bbox, mask):
    """ Crop the mask within a bounding box.

    Args:
        bbox (list) : a list of the bounding box's corners for the predicted object.
        mask : the mask of the object predicted returned by detectron2. It is actually a mask for the whole image
        and 51 requires a mask within the bbox)

    Returns: 
        A cropped mask corresponding to the inside of the bbox for the predicted object."""

    x1, y1, x2, y2 = bbox.cpu().detach().numpy().astype(np.int32)
    return mask[y1:y2, x1:x2]

def filter_confidence_labels(dataset, fo_field, confidence_thr=0):
    """Filter labels that are smaller than the confidence threshold set.

    Args:
        dataset : a :class:`fiftyone.core.dataset`
        fo_field (str): a fiftyone field with model's predictions
        confidence_thr (float) : a threshold for confidence of model's predictions 

    Returns:
        dataset : the :class:`fiftyone.core.dataset` with labels smaller than the confidence threshold removed
    """ 
    conf_filter = fo.FilterLabels(fo_field, F("confidence") > confidence_thr)
    dataset = dataset.add_stage(conf_filter)
    return dataset

def filter_small_predictions(dataset, fo_field):
    """Filter predictions covering less than 3 pixels in a dataset.

    Predictions that are very small, i.e. smaller than 3 pixels, may trigger 
    problems when exporting the fiftyone dataset. It is best practises to remove them.

    Args:
        dataset : a :class:`fiftyone.core.dataset`
        fo_field (str): a fiftyone field with model's predictions

    Returns: 
        dataset : the :class:`fiftyone.core.dataset` with small predictions area removed
    """ 
    stage = fo.FilterLabels(fo_field, F("points").length() > 0)
    dataset = dataset.add_stage(stage)
    return dataset

def predict_on_sample(sample_filepath, device, predictor, nms_threshold, classes, type_of_preds=['segm', 'bbox']):
    """Use a loaded AI model to predict segmentations mask or bounding boxes on an image.

    Args:
        sample_filepath (str): the filepath to an image (jpg, png or tif file)
        predictor : the result of build_predictor() 
        device (str): "cpu" or "cuda". Results of `"cuda" if torch.cuda.is_available() else "cpu"`
        nms_threshold (float): the non-maximum-suppression threshold.
        classes (list): a list of the classes that can be predicted by the AI model.
        type_of_preds (str): indicate if semgentation masks "segm" expected or bounding boxes "bbox"

    Returns:
        a fiftyone field :class:``fiftyone.core.detections`` containing predictions for the sample
    """
    
    image = cv2.imread(sample_filepath)
    h, w, c = image.shape

    preds = predict(predictor, sample_filepath)
    preds = apply_nms(preds, nms_threshold, type_of_preds)
    
    detections = []
    if preds:
        labels = preds['instances'].pred_classes.cpu().detach().numpy()
        scores = preds['instances'].scores.cpu().detach().numpy()
        boxes = preds['instances'].pred_boxes
        if type_of_preds=='segm':
            masks = preds['instances'].pred_masks.cpu().detach().numpy()


        if type_of_preds=='segm':

            for label, score, box, mask in zip(labels, scores, boxes, masks):
                # Convert to [top-left-x, top-left-y, width, height]
                # in relative coordinates in [0, 1] x [0, 1]
                    x1, y1, x2, y2 = box
                    rel_box = [x1 / w, y1 / h, (x2 - x1) / w, (y2 - y1) / h]

                    mask = crop_mask(box, mask)

                    detections.append(
                        fo.Detection(
                            label=classes[label],
                            bounding_box=rel_box,
                            confidence=score, 
                            mask=mask
                        )
                    )
        else:
            for label, score, box in zip(labels, scores, boxes):
                x1, y1, x2, y2 = box
                rel_box = [x1 / w, y1 / h, (x2 - x1) / w, (y2 - y1) / h]

                detections.append(
                    fo.Detection(
                    label=classes[label],
                    bounding_box=rel_box,
                    confidence=score)
                    )   
        
    return fo.Detections(detections=detections)



def add_predictions_to_dataset(dataset, predictor, device, classes, predictions_field, tags=None, type_of_preds=['segm', 'bbox'], nms_threshold=1, confidence_thr=0):
    """Add predictions to :class:`fiftyone.core.dataset`. 

    All the images :class:`fiftyone.core.dataset` that have the tag indicated will be selected to be predicted by an AI model.
    This function uses the GPU is available or parallelize through multiple CPU to predict on all the images of a dataset.

    Args:
        dataset: a :class:`fiftyone.core.dataset`  
        predictor: the result of :py:func:`artus.inference.build_predictor()` 
        device (str): "cpu" or "cuda". Results of `"cuda" if torch.cuda.is_available() else "cpu"`
        classes (list): a list of the classes predicted by the model
        predictions_field (str): the name of the fiftyone field that will be added to be dataset
        tags (list or str): a variable or a list of variables (optional) to filter the samples on which to predict
        type_of_preds (str): indicate if semgentation masks "segm" expected or bounding boxes "bbox"
        confidence_thr (float): a threshold between 0 and 1. Predictions with lower confidence than this threshold will be filtered out 
            of the :class:`fiftyone.core.dataset`

    Returns:
        a :class:`fiftyone.core.dataset` containing predictions for the tagged samples.
    """
    num_cores = 0
    detections=[]
    if tags:
        test_set = dataset.match_tags(tags)
    else:
        test_set=dataset

    with fo.ProgressBar() as pb: 
        if device == "cpu":      
            for sample in pb(test_set.values('filepath')):
                detections.append(predict_on_sample(sample, device, predictor, nms_threshold, classes, type_of_preds))

        else:
            for sample in pb(test_set.values('filepath')):
                detections.append(predict_on_sample(sample, device, predictor, nms_threshold, classes, type_of_preds))

    if type_of_preds=='segm':
        test_set.set_values("predictions_with_bbox", detections)
        
        #Transform bbox and mask in polylines
        foul.instances_to_polylines(
            test_set,
            in_field='predictions_with_bbox',
            out_field=predictions_field
            )
        
        test_set.delete_sample_field('predictions_with_bbox')

    else: 
        test_set.set_values(predictions_field, detections)

    test_set = filter_small_predictions(test_set, predictions_field)
    
    if confidence_thr>0:
        test_set = filter_confidence_labels(test_set, predictions_field, confidence_thr)

    test_set.save()
    print(test_set)

    return test_set

def apply_nms(outputs, nms_threshold, type_of_preds=['segm', 'bbox']) :
    """Apply non-maximum threshold on predictions made on a sample.

    Non Maximum Suppression NMS is a technique to select one entity, i.e. bounding boxes or masks,
    out of many overlapping entities based on the IoU (intersection over union).The function 
    applies the NMS technique for predictions of different class types, the default Detectron2 
    parameter MODEL.ROI_HEADS.NMS_THRESH_TEST applies NMS only for same class types instances
    `https://github.com/facebookresearch/detectron2/issues/978`

    Args:
        outputs (list[dict]): a list of dict in the "outputs" inference format of detectron2 lib 
        nms_threshold (float): Overlap threshold used for non-maximum suppression, suppress boxes or
            masks with IoU >= threshold
        type_of_preds (str): indicate if predictions are semgentation masks "segm" or bounding boxes "bbox"

    Returns:
        res (list[dict]): the "outputs" inference format of detectron2 lib 
    """

    detections = outputs['instances']
    pred_boxes = detections.pred_boxes.tensor
    scores = detections.scores
    pred_classes = detections.pred_classes
    if type_of_preds == 'segm':
        pred_masks = detections.pred_masks
    image_size = detections.image_size
    # Performs non-maximum suppression (NMS) on the boxes according to their intersection-over-union (IoU).
    # see :
    # https://pytorch.org/vision/stable/generated/torchvision.ops.nms.html#torchvision.ops.nms
    keep_idx = nms(pred_boxes, scores, nms_threshold)
    res = Instances(image_size)
    res.pred_boxes = Boxes(pred_boxes[keep_idx])
    res.scores = scores[keep_idx]
    res.pred_classes = pred_classes[keep_idx]
    if type_of_preds == 'segm':
        res.pred_masks = pred_masks[keep_idx]
    res = {"instances": res}

    return res

    



"""A module to train a deep learning model with a config file in the yaml format."""

from detectron2.utils.logger import setup_logger
from detectron2.data.datasets import register_coco_instances
from detectron2.config import get_cfg

import os
import torch
import yaml
from datetime import datetime

from artus.train.build_trainer import MyTrainer
from artus.train.config import add_config, read_config


def add_logs_path_to_config(config_path, checkpoint_path_dict):
    """Add logs paths to the config file to keep track of trained models.
    Args:
        config_path (str): the path to a config file (yaml)
        checkpoint_path_dict (dict): a dict including the 'CHECKPOINT' (the path to the checkpoint saved)

    Returns: 
        The same config file with the path to the final checkpoint added.
    """
    with open(config_path,'r') as config:
        config = yaml.safe_load(config) 
        config['LOGS'].update(checkpoint_path_dict)

    if config:
        with open(config_path,'w') as yamlfile:
            yaml.safe_dump(config, yamlfile) 


def train_model(config_path):
    """A function that trains a deep learning model with detectron2.

    Support multiple deep learning tasks : object detection and instance segmentation.

    Args:
        config_path : the path to a config file (yaml)

    Returns: 
    A folder within the logs directory with the checkpoint (.pth) of the model and events file 
    for visualization of the training and validation metrics with tensorboard.
    A subdirectory is also created with the inference resultst of the validation step.
    """
    setup_logger()

    device = ("cuda" if torch.cuda.is_available() else "cpu")

    #read config file for training
    config = read_config(config_path)

    #Set automatically the name of the current train session
    today = datetime.today().strftime('%Y%m%d_%H:%M:%S')
    session_name = today + os.path.basename(config_path)[:-4]

    #Register images and test/train labels
    register_coco_instances("trainset", {}, config['DATA_PATH']['COCO_TRAIN'], config['DATA_PATH']['IMAGES_DIR'])
    register_coco_instances("valset", {}, config['DATA_PATH']['COCO_VAL'], config['DATA_PATH']['IMAGES_DIR'])

    #Set the config parameters for the model
    cfg = get_cfg()
    cfg = add_config(
        cfg=cfg,
        config_path=config_path,
        output_dir=os.path.join(config['LOGS']['LOGS_DIR'], session_name),
        device=device,
        train_dataset = ("trainset",),
        test_dataset = ("valset",),
        mode='train'
        )

    os.makedirs(cfg.OUTPUT_DIR, exist_ok=True)

    #Launch training session
    trainer = MyTrainer(cfg)
    trainer.resume_or_load(resume=False)
    trainer.train()

    #Append logs information in the config file to keep track of trained model path
    checkpoint_path = os.path.join(config['LOGS']['LOGS_DIR'], session_name, 'model_final.pth')
    dict = {'CHECKPOINT' : checkpoint_path}
    add_logs_path_to_config(config_path, dict)







"""Export plots or csv after COCO-style evaluation of a trained deep learning model.

A csv file can be exported or merged to an already existed one. This csv contains 
AP, AP50, AP75, APm, APl and AP-perclass. Plots can also be exported after having exporting this csv.
The plots created are multiple histograms where each bar stands for a model.
"""

import pandas as pd
import plotly.express as px
import os

class ModelsMetricsFormat():
    """ A class that writes the evaluation results in a CSV format.
    
    If the CSV already exists, it will merge the results at the last row creating a dataframe where
    each row is one model evaluated along with their metrics.
    
    Attributes:
        session (str): the name of the directory containing a model_final.pth to evaluate
        eval_results (dict): the evaluation results from detectron2.trainer.test()
        export_dir (str): the path to the directory where metrics will be exported or
            where previous metrics.csv are saved.
        csv_name (str): a csv name to save the results of evaluation
    
    """
    def __init__(self, session, eval_results, export_dir, csv_name):
        """ Initializes the instance.

        Export_directory and csv_name are join in a complete csv_path.
        
        Args:
            session (str): the name of the directory containing a model_final.pth to evaluate
            eval_results (dict): the evaluation results from detectron2.trainer.test()
            export_dir (str): the path to the directory where metrics will be exported or
                where previous metrics.csv are saved.
            csv_name (str): a csv name to save the results of evaluation
        """
        self.session = session
        self.accuracy = eval_results
        self.csv_path = os.path.join(export_dir, csv_name)

    def check_csv_exists(self):
        """Check if a csv containing model's metrics exists or not.

        If csv already exists,results of this session will be appended to this csv. 
        Otherwise, a csv will be created.
        """
        file_exists = os.path.exists(self.csv_path)
        if file_exists:
            print('CSV containing metrics already exists. Results will be appended to this dataframe')
        else:
            print('CSV does not exist yet. We will create it for you.')
        return file_exists

    def to_pandas(self):
        """Take results from COCOevaluator and format it into a pandas dataframe."""
        segmentation_accuracies = list(self.accuracy.values())[1]
        accuracy_df = pd.DataFrame(data=segmentation_accuracies.values(), index=list(segmentation_accuracies.keys()))
        accuracy_df = accuracy_df.transpose()
        accuracy_df['model'] = self.session

        if self.check_csv_exists():
            models_metrics = pd.read_csv(self.csv_path, header=0)
            accuracy_df = pd.concat([models_metrics, accuracy_df], axis=0)

        return accuracy_df
    
    def write_to_csv(self):
        """Export results to csv format at the csv_path"""
        df = self.to_pandas()
        df.to_csv(self.csv_path, index=False)

class ModelsMetricsPlots():
    """ A class that plots the results of different models trained.

    The plots created are multiple histograms where each bar stands for a model.
    Plots are more relevant if the csv contains multiple models metrics.

    Attributes: 
        csv_metrics_path (str): the path to a csv dataframe.
        export_dir (str): the directory where the interactive plots will be exported
        plot_name (str): the name of the plot file with .html extension
        title (str): title to display on plot
    """
    
    def __init__(self, csv_metrics_path, export_dir, plot_name, title):
        """Initializes the instance.

        Args:        
            csv_metrics_path (str): the path to a csv dataframe. The results of Pÿthon(ModelsMetricsFormat.write_to_csv())
            The dataframe should have the following columns : model | AP | AP50 | AP75 | APs | APm | APl | AP-class1 | AP-class2 etc...
            One row represents the evaluation metrics of one model.
            export_dir (str): the directory where the interactive plots will be exported
            plot_name (str): the name of the plot file  with the extension (html file)
            title (str): title to display on plot
        """
        self.csv_metrics_path = csv_metrics_path
        self.export_dir = export_dir
        self.plot_name = plot_name
        self.title = title

    def process_csv(self):
        """ Read csv and pivot the dataframe.

        Create a dataframe with one row for each variable and each model.
        """
        metrics = pd.read_csv(self.csv_metrics_path, header=0)
        
        variables = metrics.columns.drop(['model'])
        
        metrics_melted = pd.melt(
            metrics, 
            id_vars='model', 
            value_vars=variables,
            value_name='Average precision')
        
        metrics_melted.dropna(axis=0, how='any', inplace=True)
        return metrics_melted

    def plot_metrics(self):
        """Plot the metrics for all the models described in the csv file.

        Plots created are facet bar plots. One facet is created per variables (AP50, Ap-class1 etc...)
        and each model precision is described by one bar.

        Returns:
            A plotly interactive barplot that can be seen with Python(fig.show())
        """
        metrics = self.process_csv()

        fig = px.bar(metrics, x="model", y="Average precision", color="model", barmode="relative",
             facet_col="variable", facet_col_wrap=5, facet_row_spacing=0.04, facet_col_spacing=0.04,
             height=1500, width = 1500, title= self.title)
        
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))

        return fig

    def export_plots(self):
        """Export the plots in html format. Plots are interactives."""
        fig = self.plot_metrics()
        fig.write_html(os.path.join(self.export_dir, self.plot_name))


    
        




"""
Read a config file and return a cfg to use a model with detectron2 library for the inference process.
"""

from detectron2 import model_zoo
from detectron2.config import get_cfg

from artus.train.config import read_config

def check_logs(config_path):
    """Check if the config file already contains path to a log directory 
    (which include a checkpoint to a model)

    Args:
        config_path (str): the path to a config file in yaml format

    Returns: 
        check_logs (bool): true if the config file includes path to a log directory, 
            false otherwise.
    """
    config = read_config(config_path)
    if config['LOGS']['CHECKPOINT']:
        check_logs = True
    else:
        check_logs = False
    return check_logs

def add_config(config_path, device):
    """ Add config for DL model coming from a yaml config file.

    The config is read and a cfg suitable for detectron2 package is returned.
    This cfg uses a model already fine tuned with detectron2. The path to this
    model'sa checkpoint should be mentionned in the LOGS section of the config file.

    Args:
        config_path (str): the path to a config file in yaml format
        device (str): 'cpu' or 'cuda'. Results of (Python("cuda" if torch.cuda.is_available() else "cpu"))
    
    Returns:
        a cfg to use a detectron model's checkpoint on unlabeled images (inference)
    """ 
    config = read_config(config_path)

    if check_logs(config_path):
        cfg = get_cfg()
        cfg.merge_from_file(model_zoo.get_config_file(config['MODEL']['URL']))
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = config['MODEL']['ROI_HEADS']['NUM_CLASSES']
        cfg.MODEL.WEIGHTS = config['LOGS']['CHECKPOINT']
        cfg.DATALOADER.NUM_WORKERS = config['DATALOADER']['NUM_WORKERS']
        cfg.SOLVER.IMS_PER_BATCH = config['SOLVER']['IMS_PER_BATCH']
        cfg.MODEL.DEVICE = device
        cfg.INPUT.MIN_SIZE_TEST = config['INPUT']['MIN_SIZE_TEST']
        cfg.INPUT.MAX_SIZE_TEST = config['INPUT']['MAX_SIZE_TEST']
    return cfg




"""A module to convert a coco file into a geojson file.

Annotations in the coco file are pixel values but if they match with a georeferenced raster
file (a tif file for example), they are spatialized with an affine transformation.

Typical usage examples:

    geojson_exporter = GeoCOCOExporter(
    coco_path = 'path/to/coco/file.json',
    sample_dir = 'path/to/tif/directory/', 
    epsg_code = 'EPSG:4326', 
    dest_path = 'path/to/export/directory/',
    dest_name = 'predictions.geojson'
)
    geojson_exporter.export()

"""

from pylabel import importer
import numpy as np
from shapely.geometry import Polygon
import solaris.vector.polygon as solpol
import geopandas
import rasterio
import os

class GeoCOCOExporter():
    """Class to convert a `coco.json` file into a geojson file. 

    Annotations in the coco file are pixel values. When converted to geojson format
    with an affine transformation, they become georeferenced.

    Attributes:
        coco_path (str): the relative or absolute path top a coco.json file
        epsg_code (str): the epsg_code (like EPSG:4326') to set the CRS of the geojson file
        dest_path (str): the destination path where the results will be exported. If the directory 
            does not exist yet, it will be created.
        dest_name (str): the name of the file to store
        sample_dir (str, optional): the path to the directory where the tif files are located. 
            If sample_name is already the full path of the sample, no need to provide sample_dir.

    """
    def __init__(self, coco_path, epsg_code, dest_path, dest_name, sample_dir=None):
        self.coco_path = coco_path
        self.epsg_code = epsg_code
        self.dest_path = dest_path
        self.dest_name = dest_name
        self.sample_dir = sample_dir

    def coco2gdf(self):
        """ Convert a coco file into a geopandas dataframe.

        Returns:
            gdf (:py:class: `geopandas.GeoDataFrame`): a dataframe with annotation masks converted into shapely polygons but still with pixel values.
                The polygons are indeed, not yet converted to spatial information.
        """

        coco_dataset = importer.ImportCoco(path=self.coco_path, name="coco_dataset")
        predictions = coco_dataset.df

        #remove empty lines in dataset, unregular polygon coordinates and nested list in annotation masks
        predictions = predictions[predictions['cat_name'] != '']
        predictions['ann_segmentation'] = [predictions['ann_segmentation'].iloc[index][0] for index in range(len(predictions))]
        predictions = predictions[predictions['ann_segmentation'].str.len() > 4]

        #convert annotation polygons (masks) into shapely polygons
        predictions['geometry'] =  [np.array_split(mask, len(mask)/2) for mask in predictions['ann_segmentation']]

        predictions['geometry'] = [Polygon(mask) for mask in predictions['geometry']]

        gdf = geopandas.GeoDataFrame(predictions)

        return gdf
    
    def get_transform(self, sample):
        """Get affine transformation for a tif sample.

        Args:
            sample (str): the path or name of the sample with the extension that should be .tif

        Returns:
            the :py:class:`affine.Affine` shapely object with the 6 elements required for an affine transformation
        """
        if self.sample_dir:
            sample = rasterio.open(os.path.join(self.sample_dir, sample))
        else:
            sample = rasterio.open(sample)
        return sample.transform
        
    def affine_transform(self, gdf):
        """ Apply an affine transformation to every polygons (annotations) in the 
        geometry column of a :py:class:`geopandas.Dataframe`

        Args:
            gdf : a g:py:class:`geopandas.Dataframe` containing a geometry column with 
                shapely polygons in pixel value.
        
        Returns:
            A :py:class:`geopandas.Dataframe` with polygons spatialized to the desire CRS
        """

        gdf['geometry'] = [solpol.convert_poly_coords(
            geom=polygon,
            affine_obj=transform) for polygon, transform in zip(gdf.geometry, gdf['transform'])]
        
        gdf = gdf.set_crs(self.epsg_code, inplace=True)

        return gdf
    
    def export(self):
        """ Export the geopandas dataframe to geojson.

        Returns:
            A geojson file exported at the dest_path + dest_name location.
        """
        #check if dest_path exists (create it if needed)
        os.makedirs(self.dest_path, exist_ok=True)

        gdf = self.coco2gdf()
        gdf['transform'] =  [self.get_transform(sample) for sample in gdf['img_filename']]
        affine_transformed_gdf = self.affine_transform(gdf)
        gdf_export = affine_transformed_gdf.filter(items=['img_filename', 'geometry', 'cat_name', 'cat_id'])
        gdf_export.to_file(self.dest_path + self.dest_name + '.geojson', driver='GeoJSON')



"""Performs COCO-style evaluation on a trained deep learning model.

To evaluate the model, a config file (yaml format) must be provided to set the model parameters.
Model is tested on the test_dataset indicated. The test set must be new to the model evaluated (not used
during training process). Results are exported in csv format at the desired location.

Typical usage examples:

  evaluate_model(config_path='/path/to/config.yml', csv_metrics_name='models_metrics.csv')
"""

from detectron2.utils.logger import setup_logger
from detectron2.data.datasets import register_coco_instances
from detectron2.config import get_cfg
from detectron2.evaluation import COCOEvaluator
import os
import torch
from artus.evaluate_model.write_eval_results import ModelsMetricsFormat
from artus.train.build_trainer import MyTrainer
from artus.train.config import read_config, add_config


def evaluate_model(config_path, csv_metrics_name):
    """ A function that reads a config file and execute a COCO evaluation on a trained model.

    Config file must be written in yaml format and must follow the example here : 
    https://github.com/6tronl/artus-examples/blob/main/configs/x101_allsites_species_overlapping25_tiles5000_ITER3000.yml
    The config file contains information needed to set the detectron2 evaluator based on 
    the model trained. The model will be evaluated using the COCOEvaluator 
    (https://detectron2.readthedocs.io/en/latest/modules/evaluation.html?highlight=COCOEvaluator#detectron2.evaluation.COCOEvaluator).
    A csv will be exported containing metrics about the model (AP50, AP, AP75, APl and AP per class). 
    If a csv already exsists in the log directory mentionned in the config file, 
    then results will be appended to it.

    Args:
         config_path (str): the path to a config file in yaml format
        csv_metrics_name (str): a CSV name to write or append the results of the evaluation.
    """
    
    setup_logger()

    #read config file for training
    config = read_config(config_path)
    
    #Register images and test/train labels
    register_coco_instances("trainset", {}, config['DATA_PATH']['COCO_TRAIN'], config['DATA_PATH']['IMAGES_DIR'])
    register_coco_instances("testset", {}, config['DATA_PATH']['COCO_TEST'], config['DATA_PATH']['IMAGES_DIR'])

    device = ("cuda" if torch.cuda.is_available() else "cpu")
    
    #Set the config parameters for the model
    cfg = get_cfg()
    cfg = add_config(
        cfg=cfg,
        config_path=config_path,
        output_dir=config['LOGS']['CHECKPOINT'],
        device=device,
        train_dataset = ("trainset",),
        test_dataset = ("testset",),
        mode='inference'
        )

    #Load model and weights
    trainer = MyTrainer(cfg)
    trainer.resume_or_load()

    #Evaluate final results
    evaluator = COCOEvaluator('testset', cfg, False, os.path.join(config['LOGS']['LOGS_DIR'], os.path.basename(config['LOGS']['CHECKPOINT'])))
    accuracy = trainer.test(cfg, trainer.model, evaluator)

    model_metrics = ModelsMetricsFormat(
        session=os.path.basename(config['LOGS']['CHECKPOINT']),
        eval_results=accuracy,
        export_dir=config['LOGS']['LOGS_DIR'],
        csv_name=csv_metrics_name
    )

    model_metrics.write_to_csv()


"""Create a fiftyone dataset with unlabeled or labeled images."""


import fiftyone as fo
import fiftyone.utils.coco as fouc
import json
import os

def create_or_load_dataset(dataset_name, dataset_type=['unlabeled', 'coco'], images_path=None, annotations_path=None, df_test=None, label_type=['segmentations', 'detections']):
    ''' Create a fiftyone labeled or unlabeled dataset.
    
    If the dataset_name provided is in the local 51 database : the dataset is loaded from the database otherwise,
    the dataset is created

    Args:
        dataset_name (str): the name of the dataset that you want to create or load
        dataset_type (str): can be 'unlabeled' for images without annotations or 'coco' if images have coco annotations.
        images_path (str): a path to the directory containing the images (can be tif, png or jpg images)
        annotations_path (str, optional): a path to the COCO annotations files
        df_test (str, optional): a path to the COCO annotations files for test images
        label_type : 'segmentations' for mask or 'detections' for bounding box annotations
    
    Returns : 
        A :class:`fiftyone.core.dataset` with at least 6 fields : id, coco_id, filepath, ground_truth annotations (if provided), tags and basic metadata.
    '''
    
    if dataset_name not in fo.list_datasets():
        if dataset_type == 'coco':
    # Create a 51 dataset with coco annotations
            importer = fouc.COCODetectionDatasetImporter(
                data_path=images_path,
                labels_path=annotations_path, 
                label_types=label_type,
                include_id=True,
                use_polylines=True,
                tolerance=2
                )
            
            dataset = fo.Dataset(
                name=dataset_name,
                persistent=True,
                overwrite=True
                )
            
            dataset.add_importer(
                dataset_importer=importer,
                label_field="ground_truth")

            dataset.persistent = True
            dataset.save()

            # Tag test samples for future usage
            if df_test:
                f = open(df_test)
                test_annot = json.load(f)

                for sample in dataset:
                    for images in test_annot['images']:
                        if os.path.basename(sample.filepath) == images['file_name']:
                            sample.tags.append('test')
                            sample.save()
        
        elif dataset_type == 'unlabeled':
            dataset = fo.Dataset.from_dir(
                name=dataset_name,
                persistent=True,
                overwrite=True,
                dataset_dir=images_path,
                dataset_type=fo.types.ImageDirectory,
                tags='test')
    
    else: 
        # load an existing dataset from the 51 database
        dataset = fo.load_dataset(dataset_name)
    
    dataset.compute_metadata()
    return dataset



"""A module to perform data augmentation during the training and/or inference process"""

from detectron2.data import detection_utils as utils
import detectron2.data.transforms as T
import copy
import torch


def custom_mapper(dataset_dict):
    """A custom mapper to make data augmentation with the images

    Custom mapper performs resizing, flipping, color editor to artificially augment data. 

    Args:
        dataset_dict (dict): a dict mapping the images along with their labels.

    Returns:
        dict: a dict mapping the augmentated images along with their labels
    """
    dataset_dict = copy.deepcopy(dataset_dict)  # it will be modified by code below
    image = utils.read_image(dataset_dict["file_name"], format="BGR")
    transform_list = [
        T.Resize((800,600)),
        T.RandomBrightness(0.8, 1.8),
        T.RandomContrast(0.6, 1.3),
        T.RandomSaturation(0.8, 1.4),
        T.RandomRotation(angle=[90, 90]),
        T.RandomLighting(0.7),
        T.RandomFlip(prob=0.4, horizontal=False, vertical=True),
    ]
    image, transforms = T.apply_transform_gens(transform_list, image)
    dataset_dict["image"] = torch.as_tensor(image.transpose(2, 0, 1).astype("float32"))

    annos = [
        utils.transform_instance_annotations(obj, transforms, image.shape[:2])
        for obj in dataset_dict.pop("annotations")
        if obj.get("iscrowd", 0) == 0
    ]
    instances = utils.annotations_to_instances(annos, image.shape[:2])
    dataset_dict["instances"] = utils.filter_empty_instances(instances)
    return dataset_dict





from rasterio.crs import CRS


def check_crs(obj, crs_code):
    """ Check if the layers provided are set to the correct CRS.

    Args:
        obj : a raster or vector layer read with rasterio or geopandas
        crs_code (str) : the CRS number to check

    Returns: 
        True or False. True means that CRS of the layer is the one expected.
    """
    crs_goal = CRS.from_epsg(crs_code)
    
    if obj.crs == crs_goal:
        print('CRS already matches with desired CRS')
        checking = True
    else:
        checking = False
    return checking
    



"""The module provides a splitter to prepare coco annotations files for deep learning training process.

The class: `COCOSplitter` splits a coco annnotations file into 3 files : the coco_train, 
coco_test and coco_val. Annotations are firstly parsed so under-reprensentated classes are removed.
Proportions of annotations per class are maintained in the splits.


Typical usage examples:

    `splitter = COCOSplitter(
        coco_path='/path/to/coco.json',
        export_dir='/path/to/export/coco/splits/',
        coco_train_name='coco_train',
        coco_test_name='coco_test',
        coco_val_name='coco_val',
        min_nb_occurences=50,
        train_pct=.8,
        val_pct=.1,
        test_pct=.1,
        batch_size=8
    )`
    `splitter.split()`

"""

from pylabel import importer, dataset
import os
from artus.evaluate_model.coco_stats import COCOStats

class COCOSplitter(COCOStats):
    """ Splits a coco file into a train, test and (optional) validation coco files. 
    
    Under-reprensentated classes are optionnally removed according the min_nb_occurrences set.
    Proportions of class annotations are kept into the splits. If no train_pct, val_pct or test_pct are set, 
    proportions of annotations in th splits will be set to 0.8, 0.1 and 0.1 respectively.

    Attributes:
        coco_path (str): a path to a coco file
        export_dir (str): a directory where the splits of the coco files will be exported
        coco_train_name (str): the name of the file with the annotations for training
        coco_test_name (str): the name of the file with the annotations for testing
        coco_val_name (str): the name of the file with the annotations for validation
        min_nb_occurrences (int): the minimum number of occurrences of a class to be 
        kept in the dataset to remove under representated classes
        train_pct (float): the fraction of annotations that will go into the train coco file
        val_pct (float): the fraction of annotations that will go into the validation coco file
        test_pct (float): the fraction of annotations that will go into the test coco file
        batch_size (int): number of images per batch process
    """

    def __init__(self, coco_path, export_dir, coco_train_name, coco_test_name, coco_val_name, min_nb_occurrences=None, train_pct=.8, val_pct=.1, test_pct=.1, batch_size=8):
        """Initializes the coco file splitter.

        Args:
            coco_path (str): a path to a coco file
            export_dir (str): a directory where the splits of the coco files will be exported
            coco_train_name (str): the name of the file with the annotations for training
            coco_test_name (str): the name of the file with the annotations for testing
            coco_val_name (str): the name of the file with the annotations for validation
            min_nb_occurrences (int): the minimum number of occurrences of a class to be 
            kept in the dataset to remove under representated classes
            train_pct (float): the fraction of annotations that will go into the train coco file
            val_pct (float): the fraction of annotations that will go into the validation coco file
            test_pct (float): the fraction of annotations that will go into the test coco file
            batch_size (int): number of images per batch process
        """        
        self.dataset = self.process_coco(coco_path, min_nb_occurrences)
        self.coco_path = coco_path
        self.export_dir = export_dir
        self.coco_train_name = coco_train_name
        self.coco_test_name = coco_test_name
        self.coco_val_name = coco_val_name
        self.train_pct = train_pct
        self.val_pct = val_pct
        self.test_pct = test_pct
        self.batch_size = batch_size

    def create_train_test_val_datasets(self):
        """Create dataset in the pylabel format to be populated by annotations aftersplitting.
        
        Returns:
            dataset_train: a training dataset in the pylabel format
            dataset_val: a validation dataset in the pylabel format
            dataset_test: a test dataset in the pylabel format
        """
        dataset_train = importer.ImportCoco(path=self.coco_path, name="trainset")
        dataset_val = importer.ImportCoco(path=self.coco_path, name="valset")
        dataset_test = importer.ImportCoco(path=self.coco_path, name="testset")
        return dataset_train, dataset_val, dataset_test
    
    def split_coco(self):
        """Splits coco files and export them in COCO format in the :py:attribute: `export_dir`.
        
        Returns:
            The proportions of class between the splits.
        """
        self.dataset.splitter.StratifiedGroupShuffleSplit(train_pct=self.train_pct, val_pct=self.val_pct, test_pct=self.test_pct, batch_size=self.batch_size)

        class_prop = self.dataset.analyze.ShowClassSplits()

        df_train = self.dataset.df.query("split == 'train'")
        df_val = self.dataset.df.query("split == 'val'")
        df_test = self.dataset.df.query("split == 'test'")
	
	
        dataset_train, dataset_val, dataset_test = self.create_train_test_val_datasets()
        
        # filter images that are moved in the other datasets
        dataset_train.df = dataset_train.df[dataset_train.df.img_filename.isin(df_train["img_filename"])].reset_index()
        # filter categories that have been deleted
        dataset_train.df = dataset_train.df[dataset_train.df.cat_name.isin(df_train["cat_name"])].reset_index()

        dataset_val.df = dataset_val.df[dataset_val.df.img_filename.isin(df_val["img_filename"])].reset_index()
        dataset_val.df = dataset_val.df[dataset_val.df.cat_name.isin(df_val["cat_name"])].reset_index()

        dataset_test.df = dataset_test.df[dataset_test.df.img_filename.isin(df_test["img_filename"])].reset_index()
        dataset_test.df = dataset_test.df[dataset_test.df.cat_name.isin(df_test["cat_name"])].reset_index()
        
        dataset_train.export.ExportToCoco(output_path=os.path.join(self.export_dir, self.coco_train_name))
        dataset_val.export.ExportToCoco(output_path=os.path.join(self.export_dir, self.coco_val_name))
        dataset_test.export.ExportToCoco(output_path=os.path.join(self.export_dir, self.coco_test_name))

        return class_prop

        
        

        



"""Build a  to train deep learning model."""

from detectron2.engine import DefaultTrainer
from detectron2.data import DatasetMapper, build_detection_train_loader, build_detection_test_loader
from detectron2.config import CfgNode
from detectron2.evaluation import COCOEvaluator
import os
from artus.train.validation_hook import LossEvalHook

class MyTrainer(DefaultTrainer):
    """A custom trainer that performs validation loop."""
    @classmethod
    def build_evaluator(cls, cfg, dataset_name, output_folder=None):
        """_summary_

        Args:
            cfg (CfgNode): config file in detectron2 format
            dataset_name (str): name of the dataset
            output_folder (str, optional): Path to the directory where results of training
                and evaluation will be saved. Defaults to None.

        Returns:
            :class:`detectron2.evaluation.COCOEvaluator`: an evaluator for coco dataset.
        """
        if output_folder is None:
            output_folder = os.path.join(cfg.OUTPUT_DIR, "inference")
        return COCOEvaluator(dataset_name, cfg, True, output_folder)

    @classmethod
    # To use the mapper inside the dataloader, you need to overwrite the build_train_loader method of the trainer
    def build_train_loader(cls, cfg):
        """Build a train dataloader.

        Args:
            cfg (CfgNode): config file in detectron2 format

        Returns:
            :class:`torch.utils.data.DataLoader` : A train dataloader.
        """
        return build_detection_train_loader(cfg)

    @classmethod
    def build_test_loader(cls, cfg:CfgNode, dataset_name):
        """Build a test dataloader.

        Args:
            cfg (CfgNode): config file in detectron2 format
            dataset_name (str): the name of the dataset

        Returns:
            :class:`torch.utils.data.DataLoader` : A test dataloader.
        """
        return build_detection_test_loader(cfg, dataset_name)

    def build_hooks(self):
        """Build a hook to make validation on the validation dataset.

        Returns:
            hooks: validation metrics
        """
        hooks = super().build_hooks()
        hooks.insert(-1,LossEvalHook(
            self.cfg.TEST.EVAL_PERIOD,
            self.model,
            build_detection_test_loader(
                self.cfg,
                self.cfg.DATASETS.TEST[0],
                DatasetMapper(self.cfg,True)
            )
        ))
        return hooks





"""Convert geojson file into coco file for deep learning purposes.

Annotations in the geojson file will be converted into pixel values with an
inverted affine transformation.
"""

import solaris.data.coco as solcoco
import os


def geojson_to_coco(tiled_geojsons, tiled_ortho, coco_dest_path, feature=None):
    """A function to convert a directory containing geojson to a single coco file.

    Args:
        tiled_geojsons (list) : a list containing paths to geojsons path, in geojson format.
        tiled_ortho (list): a list containing paths to raster tiles, in tif format.
        coco_dest_path (str): the path where the coco file will be saved. 
        feature (str): the geojson feature which is the class to affect to the polygons or bbox.

    Returns:
        A coco file saved at the coco_dest_path(.json)
    """
        
    if not os.path.exists(coco_dest_path):
        os.makedirs(coco_dest_path)
        
    coco_dict = solcoco.geojson2coco(
        image_src=tiled_ortho,
        label_src=tiled_geojsons,
        output_path=coco_dest_path,
        category_attribute=feature,
        matching_re=r'(\d+\.\d+_\-?\d+\.\d+)',
        verbose=0)



"""A module to handle a config file in yaml format and build a CfgNode.

An example of the config file expected can be found in artus's tutorials at
https://github.com/6tronl/artus-examples/blob/main/configs/x101_allsites_species_overlapping25_tiles5000_ITER3000.yml
"""

from detectron2 import model_zoo
import yaml


def read_config(config_path):
    """ Read a yaml file used to config a model for training or inference.

    Args:
        config_path (str) : the path to a config file in yaml format
    
    Returns:
        A python object 
    """
    with open(config_path) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config
	

def add_config(cfg, config_path, device, train_dataset, test_dataset, output_dir=None, mode=['train', 'inference']):
    """Add config for DL model coming from a yaml config file.

    Args:
        cfg (CfgNode): a config file in detectron2 format
        config_path (str): the path to a config file following the format provided in the tutorials.
        device (str): 'cpu' or 'cuda'. Results of `"cuda" if torch.cuda.is_available() else "cpu"`
        train_dataset (str): name of the registered dataset for training
        test_dataset (str): name of the registered dataset for validation
        output_dir (str, optional): the path to the output directory. Defaults to None.
        mode (list, optional): Whether this config will be added in a training or inference use case. 
            Defaults to ['train', 'inference'].

    Returns:
        CfgNode: a custom config file for training a deep learning model
    """
    config = read_config(config_path)
        
    cfg.merge_from_file(model_zoo.get_config_file(config['MODEL']['URL']))

    if mode == 'train':
        cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(config['MODEL']['URL'])
    elif mode == 'inference':
        cfg.MODEL.WEIGHTS = config['LOGS']['CHECKPOINT']

    cfg.DATASETS.TRAIN = train_dataset
    cfg.DATASETS.TEST = test_dataset

    cfg.TEST.EVAL_PERIOD = config['TEST']['EVAL_PERIOD']
    cfg.DATALOADER.NUM_WORKERS = config['DATALOADER']['NUM_WORKERS']
    cfg.SOLVER.IMS_PER_BATCH = config['SOLVER']['IMS_PER_BATCH']
    cfg.SOLVER.BASE_LR = config['SOLVER']['BASE_LR']
    cfg.SOLVER.MAX_ITER = config['SOLVER']['MAX_ITER']
    cfg.SOLVER.STEPS = config['SOLVER']['STEPS']
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = config['MODEL']['ROI_HEADS']['NUM_CLASSES']
    cfg.SOLVER.REFERENCE_WORLD_SIZE = 1
    cfg.MODEL_DEVICE = device
    cfg.OUTPUT_DIR = output_dir
    cfg.INPUT.MIN_SIZE_TRAIN = (config['INPUT']['MIN_SIZE_TRAIN'],)
    cfg.INPUT.MAX_SIZE_TRAIN = config['INPUT']['MAX_SIZE_TRAIN']
    cfg.INPUT.MIN_SIZE_TEST = config['INPUT']['MIN_SIZE_TEST']
    cfg.INPUT.MAX_SIZE_TEST = config['INPUT']['MAX_SIZE_TEST']
    return cfg
    
 



""" Tile raster and/or vector layer.

Depending on machine resources (such as GPU cache storage) available to user, 
the process to prepare the data for model trainings can be adapted. To process
large raster files, you can tile them into smaller tiles. Optionnally, if the 
raster is annotated with a vector layer, the vector layer can be clipped to the
same boundaries.
"""

import solaris.tile as solt
import rasterio
import os
import geopandas
import numpy as np

from artus.prepare.crs_settings import check_crs


def tile_ortho(ortho_path, dest_dir, tuple_tile_size, h_shift=0.0, v_shift=0.0):
    """ Tile a tif file.

    Tiles's size can be set with pixel values or in meters. Tiles are cut according to a regular
    grid if h_shift and v_shift are set to 0 or according to an overlapping grid. Overlaps between the tiles
    is customizable.

    Args:
        ortho_path (str): the path to raster in tif format
        dest_dir (str): the directory where the tile will be saved (create the directory if it does not exists)
        src_tile_size (tuple): a tuple indicating the length and width of the tiles produced. Can be pixel values or meters.
        h_shift (float): a float between 0 and 1 representing the fraction of shifting between horizontal neighbooring tiles
        v_shift (float): a float between 0 and 1 representing the fraction of shifting between vertical neighbooring tiles
    
    Returns:
        Tiles with 3 color channels in tif format saved in the dest_dir.
    """

    raster_tiler = solt.raster_tile.RasterTiler(
        dest_dir=dest_dir,
        src_tile_size=tuple_tile_size,
        use_src_metric_size=False,
        verbose=True,
        alpha=1,
        project_to_meters=True)

    raster_tiler.tile(
        ortho_path, 
        channel_idxs=[1,2,3])
    
    bounds = raster_tiler.tile_bounds
    
    #tile a second grid taking into account the shiftping between tiles
    if h_shift != 0.0 or v_shift!= 0.0:
        sample = rasterio.open(ortho_path)

        aoi_boundary = [
            sample.bounds.left + sample.transform[0]*(tuple_tile_size[1]*h_shift), 
            sample.bounds.bottom + sample.transform[0]*(tuple_tile_size[0]*v_shift), 
            sample.bounds.right - sample.transform[0]*(tuple_tile_size[1]*h_shift), 
            sample.bounds.top - sample.transform[0]*(tuple_tile_size[0]*v_shift)
            ]
        
        raster_tiler_shift = solt.raster_tile.RasterTiler(
            dest_dir=dest_dir,
            src_tile_size=tuple_tile_size,
            use_src_metric_size=False,
            verbose=True,
            alpha=1,
            aoi_boundary=aoi_boundary,
            project_to_meters=True)

        raster_tiler_shift.tile(
            ortho_path, 
            channel_idxs=[1,2,3])
    

        bounds = raster_tiler.tile_bounds + raster_tiler_shift.tile_bounds
    
    return bounds
    

def clip_annotated_ortho(annot_path, raster_path, matching_crs, dest_dir, tuple_tile_size, h_shift=0, v_shift=0, annot_type=['segm', 'bbox']):
    """ A function that clips a raster annotated with a vector layer to the same boundaries.
    
    Args:
        annot_path (str): the path to the vector (shapefile or geojson)
        raster_path (str): the path to an annotated raster (tif file)
        matching_crs (str): the epsg_code matching the raster and annotation file
        dest_dir (str): the path to the destination directory where tiles and matching geojsons annotations will be saved
        tuple_tile_size (tuple) : a tuple indicating the length and width of the tiles produced
        h_shift (float, optional):  a float between 0 and 1 representing the fraction of shifting between horizontal neighbooring tiles
        v_shift (float, optional): a float between 0 and 1 representing the fraction of shifting between vertical neighbooring tiles
        annot_type (str): whether the annotations are segmentation 'segm' or bounding boxes 'bbox'

    Returns: 
        A directory at the dest_dir path containig tiles and matching geojsons for each tile containing annotations.
        If the tile did not include annotations then the geojson file is not created.
    """

    ortho = rasterio.open(raster_path)
    annotations = geopandas.read_file(annot_path)

    #Make sure CRS from raster and vector layers are set to the appropriate EPSG code
    if check_crs(ortho, matching_crs) and check_crs(annotations, matching_crs):

        #remove geometry in annotations that is not a polygon
        if annot_type=='segm':
            index_to_drop = annotations.loc[annotations.geometry.geometry.type!='Polygon'].index
            annotations = annotations.drop(index=index_to_drop)

        #crop orthomosaics into tiles
        tile_bounds = tile_ortho(
            ortho_path=raster_path,
            dest_dir=dest_dir,
            tuple_tile_size=tuple_tile_size,
            h_shift=h_shift,
            v_shift=v_shift
        )

        #clip annotation layer (shp) to the same boundaries as the clipped tiles
        geojsons_dir = os.path.join(dest_dir, 'geojsons/')
        if not os.path.exists(geojsons_dir):
            os.makedirs(geojsons_dir)
            
        for tile_bound in tile_bounds:
            annotations_clipped = annotations.clip(tile_bound, keep_geom_type=True)
            if len(annotations_clipped) > 0:
                if annot_type=='segm':
                    if not annotations_clipped.loc[annotations_clipped.geometry.geometry.type!='Polygon'].empty:
                        annotations_clipped = annotations_clipped.explode()

                annotations_clipped.to_file(f'{dest_dir}/geoms_{abs(np.round(tile_bound[0], 7))}_{abs(np.round(tile_bound[3],7))}.geojson', driver='GeoJSON')





        





"""The module provides basic descriptive statisitics on a coco file (coco.json).

The COCOStats class can be used to get basic statistics on a coco file after 
deleting images without annotations in a dataset and removing under-representated
classes with the min_nb_occurrences argument.

Typical usage examples:

  `cocostats = COCOStats(coco_path='/a/path/to/coco.json', min_nb_occurrences=50)`
  `nb_images = cocostats.get_nb_images()`
  `class_occ = cocostats.get_class_stats()`
  `cocostats.export_stats(export_path='/path/to/dir/export.csv)`
"""

from pylabel import importer

def rm_min_classes(dataset, min_nb_occurrences):
    """Remove classes that do not reach the minimum number of occurences per class.

    Allows to get rid of under-representated in a dataset before splitting a coco file
    for deep learning purposes.

    Args:
        dataset: a coco file read with :func:`importer.ImportCoco()` from pylabel. This function
        returns a dataframe where one row is one label
        min_nb_occurrences (int): an integer that is the minimum number of occurrences 
        of a class to be kept in the dataset

    Returns:
        The same dataframe format but classes that have less than the minimum number of 
        occurrences are removed.
    """
    grouped_by_class = dataset.df.groupby(by='cat_name', axis=0).count()
    under_represented_classes = grouped_by_class[grouped_by_class['img_folder'] < min_nb_occurrences].index
    deleted_classes=list(under_represented_classes.values)
    index_to_remove = dataset.df[dataset.df.cat_name.isin(deleted_classes)].index
    dataset.df.drop(index=index_to_remove, inplace=True)
    return dataset

def rm_tiles_without_annot(dataset):
    """Remove tiles without annotations.

    Args:
        dataset: a coco file read with importer.ImportCoco() from pylabel. This function
        returns a dataframe where one row is one label

    Returns:
        The same dataframe format but rows with empty column 'cat_name' are removed.
    """
    ind_images_without_annot = dataset.df.loc[dataset.df['cat_name']==''].index
    dataset.df.drop(index=ind_images_without_annot, inplace=True)
    return dataset


class COCOStats():
    """A class to explore the basic stats of a coco file.

    The COCOStats class can be used to get basic statistics on a coco file after 
    deleting images without annotations in a dataset and removing under-representated
    classes with the min_nb_occurrences argument. Statistics can be printed or exported 
    in a csv file.

    Attributes:
        coco_path (str): a path to a coco file
        min_nb_occurrences (int): an integer that is the minimum number of occurrences of a 
        class to be kept in the dataframe.
    """
    def __init__(self, coco_path, min_nb_occurrences=None):
        """Initializes the instance by reading the coco_path file and 
            removing under-representated class if min_nb_occurrences!=None.

        Args:
            coco_path (str): a path to a coco file
            min_nb_occurrences (int): an integer that is the minimum number of occurrences of a 
            class to be kept in the dataframe, remove under representated classes
        """
        self.dataset = self.process_coco(coco_path, min_nb_occurrences)

    def process_coco(self, coco_path, min_nb_occurrences):
        """Read coco file and removes the samples without annotations and underrepresetned classes if needed.
        
        Args:
            coco_path (str): a path to a coco file
            min_nb_occurrences (int): an integer that is the minimum number of occurrences of a 
            class to be kept in the dataframe, remove under representated classes.
        
        Returns:
            A dataframe where one row is one label.
        """        
        dataset = importer.ImportCoco(path=coco_path, name="dataset")  
        dataset = rm_tiles_without_annot(dataset)
        if min_nb_occurrences:
            dataset = rm_min_classes(dataset, min_nb_occurrences)
        return dataset
    
    def get_class_stats(self):
        """Print the number of occurrences per classes and number of classes in dataframe
            after removing images without annotations and under-representated classes."""
        print(f"Classes:{self.dataset.analyze.classes}")
        print(f"Number of classes: {self.dataset.analyze.num_classes}")
        print(f"Class counts:\n{self.dataset.analyze.class_counts}")
        
    def get_nb_images(self):
        """Print the number of images in dataframe after removing images without 
            annotations and under-representated classes."""
        print(f"Number of images: {self.dataset.analyze.num_images}")
    
    def export_stats(self, export_path):
        """ Export the number of occurrences per classin csv format.
         
        Statistics are exported after removing images without annotations and (optional) under-representated classes.
        
        Args:
            export_path (str): csv path to export the class counts.
        
        """
        self.dataset.analyze.class_counts.to_csv(export_path)

